package ch.zhaw.pm2.racetrack;

import ch.zhaw.pm2.racetrack.given.ConfigSpecification;
import ch.zhaw.pm2.racetrack.given.TrackSpecification;

import java.io.File;
import java.io.FileNotFoundException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Scanner;

/**
 * This class represents the racetrack board.
 *
 * <p>The racetrack board consists of a rectangular grid of 'width' columns and 'height' rows.
 * The zero point of he grid is at the top left. The x-axis points to the right and the y-axis points downwards.</p>
 * <p>Positions on the track grid are specified using {@link PositionVector} objects. These are vectors containing an
 * x/y coordinate pair, pointing from the zero-point (top-left) to the addressed space in the grid.</p>
 *
 * <p>Each position in the grid represents a space which can hold an enum object of type {@link Config.SpaceType}.<br>
 * Possible Space types are:
 * <ul>
 *  <li>WALL : road boundary or off track space</li>
 *  <li>TRACK: road or open track space</li>
 *  <li>FINISH_LEFT, FINISH_RIGHT, FINISH_UP, FINISH_DOWN :  finish line spaces which have to be crossed
 *      in the indicated direction to winn the race.</li>
 * </ul>
 * <p>Beside the board the track contains the list of cars, with their current state (position, velocity, crashed,...)</p>
 *
 * <p>At initialization the track grid data is read from the given track file. The track data must be a
 * rectangular block of text. Empty lines at the start are ignored. Processing stops at the first empty line
 * following a non-empty line, or at the end of the file.</p>
 * <p>Characters in the line represent SpaceTypes. The mapping of the Characters is as follows:</p>
 * <ul>
 *   <li>WALL : '#'</li>
 *   <li>TRACK: ' '</li>
 *   <li>FINISH_LEFT : '&lt;'</li>
 *   <li>FINISH_RIGHT: '&gt;'</li>
 *   <li>FINISH_UP   : '^;'</li>
 *   <li>FINISH_DOWN: 'v'</li>
 *   <li>Any other character indicates the starting position of a car.<br>
 *       The character acts as the id for the car and must be unique.<br>
 *       There are 1 to {@link Config#MAX_CARS} allowed. </li>
 * </ul>
 *
 * <p>All lines must have the same length, used to initialize the grid width).
 * Beginning empty lines are skipped.
 * The the tracks ends with the first empty line or the file end.<br>
 * An {@link InvalidTrackFormatException} is thrown, if
 * <ul>
 *   <li>not all track lines have the same length</li>
 *   <li>the file contains no track lines (grid height is 0)</li>
 *   <li>the file contains more than {@link Config#MAX_CARS} cars</li>
 * </ul>
 *
 * <p>The Track can return a String representing the current state of the race (including car positons)</p>
 */
public class Track implements TrackSpecification {

    public static final char CRASH_INDICATOR = 'X';

    // TODO: Add necessary variables
    private List<String> track;
    private List<Car> cars;

    /**
     * Initialize a Track from the given track file.
     *
     * @param trackFile Reference to a file containing the track data
     * @throws FileNotFoundException       if the given track file could not be found
     * @throws InvalidTrackFormatException if the track file contains invalid data (no tracklines, ...)
     */
    public Track(File trackFile) throws FileNotFoundException, InvalidTrackFormatException {
        // TODO:
        track = new ArrayList<>();
        cars = new ArrayList<>();
        readFile(trackFile);

        //TODO: throw error again...
        //throw new UnsupportedOperationException();    // was allready in !!
    }

    /**
     * This method reads the File and saves it to the track ArrayList Line by Line
     *
     * @param trackFile the File where the track has been documented
     * @throws FileNotFoundException if the FilePath is invalid.
     */
    private void readFile(File trackFile) throws FileNotFoundException {
        Scanner scanner = new Scanner(trackFile);
        while (scanner.hasNextLine()) {
            track.add(scanner.nextLine());
        }
    }

    public void addCar() {
        ConfigSpecification.SpaceType[] spaceTypes = ConfigSpecification.SpaceType.values();
        ArrayList<Character> allSpaceTypesAsChar = new ArrayList<>();

        for (ConfigSpecification.SpaceType spaceType : spaceTypes) {
            allSpaceTypesAsChar.add(spaceType.getValue());
        }


        for (String line : track) {
            for (int i = 0; i < line.length(); i++) {
                if (!allSpaceTypesAsChar.contains(line.charAt(i))) {
                    //TODO: CREATE CAR - ADD IT TO CARS - BUT CHECK THAT Symbol is not allready used for another car?
                }
            }
        }

    }

    private PositionVector findChar(char symbol) {
        PositionVector vector = null;
        for (int i = 0; i < track.size(); i++) {
            String line = track.get(i);
            for (int j = 0; j < line.length(); j++) {
                if (line.charAt(j) == symbol) {
                    vector = new PositionVector(j,i);
                }
            }
        }
        return vector;
    }

    private void findFinish() {
        //TODO:

    }


    /**
     * @return the track
     */
    public List<String> getTrack() {
        return track;
    }

    /**
     * Return the type of space at the given position.
     * If the location is outside the track bounds, it is considered a wall.
     *
     * @param position The coordinates of the position to examine
     * @return The type of track position at the given location
     */
    @Override
    public Config.SpaceType getSpaceType(PositionVector position) {
        // TODO: implementation // Done first tests passed but what to do with error
        char charAtPosition = track.get(position.getY()).charAt(position.getX());

        ConfigSpecification.SpaceType[] spaceTypes = ConfigSpecification.SpaceType.values();
        for (ConfigSpecification.SpaceType spaceType : spaceTypes) {
            if (spaceType.getValue() == charAtPosition) {
                return spaceType;
            }
        }

        throw new UnsupportedOperationException();
    }

    /**
     * Return the number of cars.
     *
     * @return Number of cars
     */
    @Override
    public int getCarCount() {
        // TODO: error???
        return cars.size();
        //throw new UnsupportedOperationException();
    }

    /**
     * Get instance of specified car.
     *
     * @param carIndex The zero-based carIndex number
     * @return The car instance at the given index
     */
    @Override
    public Car getCar(int carIndex) {
        // TODO: error???
        return cars.get(carIndex);
        //throw new UnsupportedOperationException();
    }

    /**
     * Get the id of the specified car.
     *
     * @param carIndex The zero-based carIndex number
     * @return A char containing the id of the car
     */
    @Override
    public char getCarId(int carIndex) {
        // TODO: error???
        return cars.get(carIndex).getID();
        //throw new UnsupportedOperationException();
    }

    /**
     * Get the position of the specified car.
     *
     * @param carIndex The zero-based carIndex number
     * @return A PositionVector containing the car's current position
     */
    @Override
    public PositionVector getCarPos(int carIndex) {
        // TODO: nextPosition or Position of Trackfile?
        //Alternative 1
        return cars.get(carIndex).nextPosition();

        //Alternative 2 //NULL aussschliessen falls umgesetzt.
        return findChar(cars.get(carIndex).getID());

        //throw new UnsupportedOperationException();
    }

    /**
     * Get the velocity of the specified car.
     *
     * @param carIndex The zero-based carIndex number
     * @return A PositionVector containing the car's current velocity
     */
    @Override
    public PositionVector getCarVelocity(int carIndex) {
        // TODO: error???
        return cars.get(carIndex).getVelocity();
        //throw new UnsupportedOperationException();
    }

    /**
     * Gets character at the given position.
     * If there is a crashed car at the position, {@link #CRASH_INDICATOR} is returned.
     *
     * @param y            position Y-value
     * @param x            position X-vlaue
     * @param currentSpace char to return if no car is at position (x,y)
     * @return character representing position (x,y) on the track
     */
    @Override
    public char getCharAtPosition(int y, int x, Config.SpaceType currentSpace) {
        // TODO: implementation
        throw new UnsupportedOperationException();
    }

    /**
     * Return a String representation of the track, including the car locations.
     *
     * @return A String representation of the track
     */
    @Override
    public String toString() {
        // TODO: error???
        String str = "";
        for (String line : track) {
            str += line+"\n";
        }
        return str;
        //throw new UnsupportedOperationException();
    }
}
